'''Parse binary data painless describing the packets' structures in python classes: no need to write 'for' loops or nested 'if' conditionals.'''

__version__ = "0.5.0"

_author  = 'Gehn'
_license = 'GNU LGPLv3'
_url = 'https://github.com/gehnch/bisturi'
